/* Copyright (c) 2015 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/** @file mqtt_rx.c
 *
 * @brief Handles packet receive on transport TCP or TLS.
 */
#include "mqtt_internal.h"

void event_notify(mqtt_client_t * const p_client, const mqtt_evt_t * p_evt, uint32_t flags);
void disconnect_event_notify(mqtt_client_t * p_client, uint32_t result);


static uint32_t mqtt_handle_packet(mqtt_client_t * p_client,
                                   uint8_t       * p_data,
                                   uint32_t        datalen,
                                   uint32_t        offset)
{
    mqtt_evt_t    evt;
    uint32_t      err_code = NRF_SUCCESS;
    bool          notify_event = true;
    
    // Success by default, overwritten in special cases.
    evt.result = NRF_SUCCESS;
    
    switch (p_data[0] & 0xF0)
    {
        case MQTT_PKT_TYPE_CONNACK:
        {
            MQTT_TRC("[MQTT]:[%p]: Received MQTT_PKT_TYPE_CONNACK!\r\n", p_client);

            if (p_client->protocol_version == MQTT_VERSION_3_1_1)
            {
                evt.param.connack.session_present_flag = p_data[2] & MQTT_HEADER_CONNACK_MASK;

                MQTT_TRC("[MQTT]:[%p]: session_present_flag: %d\r\n",
                         p_client,
                         evt.param.connack.session_present_flag);
            }

            evt.param.connack.return_code = 
                                   (mqtt_conn_return_code_t)(p_data[3] & MQTT_HEADER_CONNACK_MASK);

            MQTT_TRC("[MQTT]:[%p]: return_code:          %d\r\n",
                     p_client,
                     evt.param.connack.return_code);

            if (evt.param.connack.return_code == MQTT_CONNECTION_ACCEPTED)
            {
                // Set state.
                MQTT_SET_STATE(p_client, MQTT_STATE_CONNECTED);
            }

            evt.result = evt.param.connack.return_code;
            evt.id = MQTT_EVT_CONNACK;

            break;
        }
        case MQTT_PKT_TYPE_PUBLISH:
        {
            evt.param.publish.dup_flag          = p_data[0] & MQTT_HEADER_DUP_MASK;
            evt.param.publish.retain_flag       = p_data[0] & MQTT_HEADER_RETAIN_MASK;
            evt.param.publish.message.topic.qos = ((p_data[0] & MQTT_HEADER_QOS_MASK) >> 1);

            MQTT_TRC("[MQTT]:[CID %p]: Received MQTT_PKT_TYPE_PUBLISH, QoS:%02x\r\n",
                     p_client, evt.param.publish.message.topic.qos);

            err_code = unpack_utf8_str(&evt.param.publish.message.topic.topic,
                                       datalen,
                                       p_data,
                                       &offset);

            if (err_code == NRF_SUCCESS)
            {
                if (evt.param.publish.message.topic.qos)
                {
                    err_code = unpack_uint16(&evt.param.publish.message_id,
                                             datalen,
                                             p_data,
                                             &offset);
                }
            }

            if (err_code == NRF_SUCCESS)
            {
                err_code = unpack_bin_str(&evt.param.publish.message.payload,
                                          datalen,
                                          p_data,
                                          &offset);

                // Zero length publish messages are permitted.
                if (err_code != NRF_SUCCESS)
                {
                     evt.param.publish.message.payload.p_bin_str = NULL;
                     evt.param.publish.message.payload.bin_strlen = 0;
                }
            }

            MQTT_TRC("[MQTT]: PUB message len %08x, topic len %08x\r\n",
                     evt.param.publish.message.payload.bin_strlen,
                     evt.param.publish.message.topic.topic.utf_strlen);

            evt.id     = MQTT_EVT_PUBLISH;
            evt.result = err_code;

            UNUSED_VARIABLE(iot_timer_wall_clock_get(&p_client->last_activity));

            break;
        }
        
        case MQTT_PKT_TYPE_PUBACK:
        {
            MQTT_TRC("[MQTT]: Received MQTT_PKT_TYPE_PUBACK!\r\n");
            
            evt.id     = MQTT_EVT_PUBACK;            
            err_code   = unpack_uint16(&evt.param.puback.message_id,
                                       datalen,
                                       p_data,
                                       &offset);
            evt.result = err_code;            
            break;
        }
        
        case MQTT_PKT_TYPE_PUBREC:
        {
            MQTT_TRC("[MQTT]: Received MQTT_PKT_TYPE_PUBREC!\r\n");
            
            evt.id     = MQTT_EVT_PUBREC;
            err_code   = unpack_uint16(&evt.param.pubrec.message_id,
                                       datalen,
                                       p_data,
                                       &offset);
            evt.result = err_code;                        
            break;
        }
        case MQTT_PKT_TYPE_PUBREL:
        {
            MQTT_TRC("[MQTT]: Received MQTT_PKT_TYPE_PUBREL!\r\n");
            
            evt.id     = MQTT_EVT_PUBREL;
            err_code   = unpack_uint16(&evt.param.pubrel.message_id,
                                       datalen,
                                       p_data,
                                       &offset);
            evt.result = err_code;
            break;
        }
        case MQTT_PKT_TYPE_PUBCOMP:
        {
            MQTT_TRC("[MQTT]: Received MQTT_PKT_TYPE_PUBCOMP!\r\n");
            
            evt.id     = MQTT_EVT_PUBCOMP;
            err_code   = unpack_uint16(&evt.param.pubcomp.message_id,
                                       datalen,
                                       p_data,
                                       &offset);
            evt.result = err_code;
            break;
        }
        case MQTT_PKT_TYPE_SUBACK:
        {
            MQTT_TRC("[MQTT]: Received MQTT_PKT_TYPE_SUBACK!\r\n");
            
            evt.id     = MQTT_EVT_SUBACK;
            err_code   = unpack_uint16(&evt.param.pubrec.message_id,
                                       datalen,
                                       p_data,
                                       &offset);
            evt.result = err_code;            
            break;
        }
        case MQTT_PKT_TYPE_UNSUBACK:
        {
            MQTT_TRC("[MQTT]: Received MQTT_PKT_TYPE_UNSUBACK!\r\n");
            
            evt.id     = MQTT_EVT_UNSUBACK;
            err_code   = unpack_uint16(&evt.param.pubrec.message_id,
                                       datalen,
                                       p_data,
                                       &offset);
            evt.result = err_code;            
            break;
        }
        case MQTT_PKT_TYPE_PINGRSP:
        {
            MQTT_TRC("[MQTT]: Received MQTT_PKT_TYPE_PINGRSP!\r\n");
            
            // No notification of Ping response to application.
            notify_event = false;
            break;
        }
        default:
        {
            // Nothing to notify.
            notify_event = false;
            break;
        }
    }
    
    if (notify_event == true)
    {
        event_notify(p_client, &evt, MQTT_EVT_FLAG_NONE);
    }

    return err_code;
}


uint32_t mqtt_handle_rx_data(mqtt_client_t * p_client, uint8_t * p_data, uint32_t datalen)
{
    uint32_t      err_code = NRF_SUCCESS;
    uint32_t      offset = 0;

    while (offset < datalen)
    {
        uint32_t start = offset;
        uint32_t remaining_length = 0;

        offset = 1; // Skip first byte to offset MQTT packet length.
        err_code = packet_length_decode(p_data + start,
                                        datalen - start,
                                        &remaining_length,
                                        &offset);
        if (err_code != NRF_SUCCESS)
        {
            return err_code;
        }

        uint32_t packet_length = offset + remaining_length;

        if (start + packet_length > datalen)
        {
            return NRF_ERROR_INVALID_LENGTH;
        }

        err_code = mqtt_handle_packet(p_client,
                                      p_data + start,
                                      packet_length,
                                      offset);
        if (err_code != NRF_SUCCESS)
        {
            return err_code;
        }

        offset = start + packet_length;
    }

    return err_code;
}
