/* Copyright (c) 2016 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/** @file mqtt_transport.h
 *
 * @brief Internal functions to handle transport in MQTT module.
 */

#ifndef MQTT_TRANSPORT_H_
#define MQTT_TRANSPORT_H_

#include "mqtt.h"


/**@brief Transport for handling transport connect procedure. */
typedef uint32_t (*transport_connect_handler_t)(mqtt_client_t * p_client);

/**@brief Transport write handler. */
typedef uint32_t (*transport_write_handler_t)(mqtt_client_t * p_client, uint8_t const * data, uint32_t datalen);

/**@brief Transport read handler. */
typedef uint32_t (*transport_read_handler_t)(mqtt_client_t * p_client, uint8_t * data, uint32_t datalen);

/**@brief Transport disconnect handler. */
typedef uint32_t (*transport_disconnect_handler_t)(mqtt_client_t * p_client);

/**@brief Transport procedure handlers. */
typedef struct
{
    transport_connect_handler_t    connect;                                                         /**< Transport connect handler. Handles TCP connection callback based on type of transport.*/
    transport_write_handler_t      write;                                                           /**< Transport write handler. Handles transport write based on type of transport. */
    transport_read_handler_t       read;                                                            /**< Transport read handler. Handles transport read based on type of transport. */
    transport_disconnect_handler_t disconnect;                                                      /**< Transport disconnect handler. Handles transport disconnection based on type of transport. */
} transport_procedure_t;


/**@brief Handles TCP Connection Complete for configured transport.
 *
 * @param[in] p_client Identifies the client on which the procedure is requested.
 *
 * @retval NRF_SUCCESS or an error code indicating reason for failure.
 */
uint32_t mqtt_transport_connect(mqtt_client_t * p_client);


/**@brief Handles write requests on configured transport.
 *
 * @param[in] p_client Identifies the client on which the procedure is requested.
 * @param[in] p_data   Data to be written on the transport.
 * @param[in] datalen  Length of data to be written on the transport.
 *
 * @retval NRF_SUCCESS or an error code indicating reason for failure.
 */
uint32_t mqtt_transport_write(mqtt_client_t * p_client, uint8_t const * p_data, uint32_t datalen);


/**@brief Handles read requests on configured transport.
 *
 * @param[in] p_client Identifies the client on which the procedure is requested.
 * @param[in] p_data   Pointer where read data is to be fetched.
 * @param[in] datalen  Size of memory provided for the operation.
 *
 * @retval NRF_SUCCESS or an error code indicating reason for failure.
 */
uint32_t mqtt_transport_read(mqtt_client_t * p_client, uint8_t * p_data, uint32_t datalen);


/**@brief Handles transport disconnection requests on configured transport.
 *
 * @param[in] p_client Identifies the client on which the procedure is requested.
 *
 * @retval NRF_SUCCESS or an error code indicating reason for failure.
 */
uint32_t mqtt_transport_disconnect(mqtt_client_t * p_client);


/**@brief Initiates TCP Connection.
 *
 * @param[in] p_client Identifies the client on which the procedure is requested.
 *
 * @retval NRF_SUCCESS or an error code indicating reason for failure.
 */
uint32_t tcp_request_connection(mqtt_client_t * p_client);


/**
 * @brief Wait for an incoming MQTT packet.
 *        The registered callback will be called with the packet payload.
 *
 * @param[in]  p_client  Client instance for which the procedure is requested.
 *                       Shall not be NULL.
 * @param[in]  timeout   Maximum interval (in milliseconds) to wait for a packet.
 *                       If timeout is 0, the interval is indefinitely.
 *
 * @retval NRF_SUCCESS or an error code indicating reason for failure.
 */
uint32_t tcp_receive_packet(mqtt_client_t * p_client, uint32_t timeout);


/**@brief Handles TCP Connection Complete for TCP(non-secure) transport.
 *
 * @param[in] p_client Identifies the client on which the procedure is requested.
 *
 * @retval NRF_SUCCESS or an error code indicating reason for failure.
 */
uint32_t mqtt_client_tcp_connect(mqtt_client_t * p_client);


/**@brief Handles write requests on TCP(non-secure) transport.
 *
 * @param[in] p_client Identifies the client on which the procedure is requested.
 * @param[in] p_data   Data to be written on the transport.
 * @param[in] datalen  Length of data to be written on the transport.
 *
 * @retval NRF_SUCCESS or an error code indicating reason for failure.
 */
uint32_t mqtt_client_tcp_write(mqtt_client_t * p_client, uint8_t const * p_data, uint32_t datalen);


/**@brief Handles read requests on TCP(non-secure) transport.
 *
 * @param[in] p_client Identifies the client on which the procedure is requested.
 * @param[in] p_data   Pointer where read data is to be fetched.
 * @param[in] datalen  Size of memory provided for the operation.
 *
 * @retval NRF_SUCCESS or an error code indicating reason for failure.
 */
uint32_t mqtt_client_tcp_read(mqtt_client_t * p_client, uint8_t * p_data, uint32_t datalen);


/**@brief Handles transport disconnection requests on TCP(non-secure) transport.
 *
 * @param[in] p_client Identifies the client on which the procedure is requested.
 *
 * @retval NRF_SUCCESS or an error code indicating reason for failure.
 */
uint32_t mqtt_client_tcp_disconnect(mqtt_client_t * p_client);


/**@brief Handles read requests on TLS(secure) transport.
 *
 * @param[in] p_client Identifies the client on which the procedure is requested.
 * @param[in] p_data   Pointer where read data is to be fetched.
 * @param[in] datalen  Size of memory provided for the operation.
 *
 * @retval NRF_SUCCESS or an error code indicating reason for failure.
 */
uint32_t mqtt_client_tls_connect(mqtt_client_t * p_client);


/**@brief Handles write requests on TLS(secure) transport.
 *
 * @param[in] p_client Identifies the client on which the procedure is requested.
 * @param[in] p_data   Data to be written on the transport.
 * @param[in] datalen  Length of data to be written on the transport.
 *
 * @retval NRF_SUCCESS or an error code indicating reason for failure.
 */
uint32_t mqtt_client_tls_write(mqtt_client_t * p_client, uint8_t const * p_data, uint32_t datalen);


/**@brief Handles read requests on TLS(secure) transport.
 *
 * @param[in] p_client Identifies the client on which the procedure is requested.
 * @param[in] p_data   Pointer where read data is to be fetched.
 * @param[in] datalen  Size of memory provided for the operation.
 *
 * @retval NRF_SUCCESS or an error code indicating reason for failure.
 */
uint32_t mqtt_client_tls_read(mqtt_client_t * p_client, uint8_t * p_data, uint32_t datalen);


/**@brief Handles transport disconnection requests on TLS(secure) transport.
 *
 * @param[in] p_client Identifies the client on which the procedure is requested.
 *
 * @retval NRF_SUCCESS or an error code indicating reason for failure.
 */
uint32_t mqtt_client_tls_disconnect(mqtt_client_t * p_client);


/**@brief Aborts TCP connection.
 *
 * @param[in] p_client Identifies the client on which the procedure is requested.
 *
 * @retval NRF_SUCCESS or an error code indicating reason for failure.
 */
void mqtt_client_tcp_abort(mqtt_client_t * p_client);


#endif // MQTT_TRANSPORT_H_
